import uuid
from django.db import models

# Create your models here.


class CreatedUpdatedModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class TaskExecution(CreatedUpdatedModel):

    # APPLIED = 'APPLIED'
    # REJECTED = 'REJECTED'

    # USER_CHOICES = (
    #     (APPLIED, APPLIED),
    #     (REJECTED, REJECTED),
    # )


    # choice = models.CharField(null=False, max_length=32, blank=False, choices=USER_CHOICES)

    # id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    task_name = models.CharField(null=False, max_length=256, blank=False)
    task_path = models.CharField(null=False, max_length=256, blank=False)

    task_args = models.JSONField(null=False, blank=False, default=list)
    task_kwargs = models.CharField(null=False, blank=False, default=dict)

    asked_at = models.DateTimeField(null=False, blank=True)
    started_at = models.DateTimeField(null=False, blank=True)
    finished_at = models.DateTimeField(null=False, blank=True)




from collections import namedtuple
# from typing import NamedTuple

# class Person(NamedTuple):
#     name: str
#     age: int
#     height: float
#     weight: float
#     country: str = "Canada"


Queue = namedtuple('Queue', [
    'queue_name',
    'subscription_name',
    'push_endpoint'
])


Schedule = namedtuple('Schedule', [
    'module_name',
    'func_name',
    'cron',
    'args',
    'kwargs',
])



# I3TASKS = {
#     "NAMESPACE": f"tasks.{SHORT_PROJECT_NAME}",
#     'QUEUES': {
#         'DEFAULT_QUEUE': {
#             "QUEUE_NAME": 'default',
#             "SUBSCRIPTION_NAME": 'default',
#             "PUSH_ENDPOINT": "http://pwd-backend:9577/i3/tasks-push/",
#         },
#         'OTHER_QUEUES': [

#         ]
#     },
#     'SCHEDULES': [
#         {
#             'module_name': 'i3tasks.tasks',
#             'func_name': 'test_task',
#             'cron': '* * * * *',
#             'args': [],
#             'kwargs': {},
#         },
#     ]
# }

# I3TasksSettings = namedtuple('I3TasksSettings', [
#     'namespace',
#     'queues',
#     'schedules'
# ])


class I3TasksSettings():
    namespace = 'itasks'
    default_queue = None
    other_queues = []
    schedules = []

    def __init__(
            self,
            namespace: str,
            default_queue: Queue,
            other_queues: 'tuple[Queue, ...]',
            schedules: 'tuple[Schedule, ...]',
        ) -> None:
        self.namespace = namespace
        self.default_queue = default_queue
        self.other_queues = other_queues
        self.schedules = schedules


from django.contrib import admin

from .models import TaskExecution


@admin.register(TaskExecution)
class TaskExecutionAdmin(admin.ModelAdmin):
    list_display = [
        "task_name",
        "task_path",
        "asked_at",
        "started_at",
        "finished_at",
    ]



import logging

from .utils import TaskDecorator


logger = logging.getLogger(__name__)


@TaskDecorator
def test_task(*args, **kwargs):
    mex = f"This is the test task with args: {args} and kwargs: {kwargs}"
    logger.info(mex)
    return mex



# from django.shortcuts import render

# Create your views here.
import json
import logging
import base64
import datetime
import importlib
from django.views.decorators.csrf import csrf_exempt

# from django.views.decorators.http import require_POST
from django.conf import settings

from croniter import croniter

from django.http import HttpResponse
from django.http import JsonResponse

from django.views import View

from django.utils.decorators import method_decorator

from django_i3tasks.utils import TaskDecorator


logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name='dispatch')
class BeatTaskView(View):

    def import_task(self, meta_info: dict) -> TaskDecorator:
        task = None
        try:
            # equiv. of your `import matplotlib.text as text`
            task_module = importlib.import_module(meta_info["module_name"])
            task = getattr(task_module, meta_info["func_name"])
        except Exception as exc:
            logging.error(exc)
            raise
        return task

    @csrf_exempt
    def post(self, request, *args, **kwargs):
        # if not request.body:
        #     return HttpResponse("missing post data", status=400)

        # try:
        #     data = self.get_data(request_body=request.body)
        # except Exception:
        #     HttpResponse("impossible to decode body", status=400)

        # if data is None:
        #     return HttpResponse("None Data", status=400)
        i3tasks_settings = getattr(settings, 'I3TASKS', {})
        schedules = getattr(i3tasks_settings, 'schedules', [])

        now = datetime.datetime.utcnow()

        logger.info("test task matching cron at %s", now.isoformat())
        for schedule in schedules:
            try:
                module_name = schedule.module_name
                func_name = schedule.func_name
                _args = schedule.args
                _kwargs = schedule.kwargs

                cron = schedule.cron

                cron_is_matched = croniter.match(cron, now)
                logger.info(
                    "test task matching %s.%s at %s for %s",
                    module_name,
                    func_name,
                    now,
                    cron
                )
                if cron_is_matched:
                    logger.info(
                        "launching async task %s.%s at %s",
                        module_name,
                        func_name,
                        datetime.datetime.utcnow().isoformat()
                    )
                    task = self.import_task(meta_info={
                        'module_name': module_name,
                        'func_name': func_name
                    })
                    task.async_run(*_args, **_kwargs)
                    # task_module = importlib.import_module(module_name)
                    # task = getattr(task_module, func_name)
                else:
                    iter = croniter(cron, now)
                    print(iter.get_next(datetime))

            except Exception as exc:
                logger.error(exc)

        return JsonResponse({'status': 'ok'}, status=200)


@method_decorator(csrf_exempt, name='dispatch')
class PushedTaskView(View):
    def get_data(self, request_body):
        data = None

        try:
            json_body = json.loads(request_body)
            logger.info(f"json_body ==> {json_body}")
            logger.info(
                f'encoding ==> {json_body["message"]["attributes"]["encoding"]}'
            )
            logger.info(f'data ==> {json_body["message"]["data"]}')
            # decoded_data = str(
            #     object=json_body["message"]["data"],
            #     encoding=json_body["message"]["attributes"]["encoding"],
            #     errors="strict",
            # )
            decoded_data_base4 = base64.b64decode(json_body["message"]["data"])
            logger.info(f"data decoded base64 ==> {decoded_data_base4}")
            decoded_data = decoded_data_base4.decode(
                encoding=json_body["message"]["attributes"]["encoding"],
                errors="strict",
            )
            # logger.info(
            logger.info(f"data decoded ==> {decoded_data}")
            data = json.loads(decoded_data)
            logger.info(f"data ==> {data}")

        except Exception as exc:
            logger.error(exc)
            raise

        return data

    def import_task(self, meta_info):
        task = None
        try:
            # equiv. of your `import matplotlib.text as text`
            task_module = importlib.import_module(meta_info["module_name"])
            task = getattr(task_module, meta_info["func_name"])
        except Exception as exc:
            logging.error(exc)
            raise
        return task

    @csrf_exempt
    def post(self, request, *args, **kwargs):
        if not request.body:
            logger.error(f"missing post data")
            return JsonResponse({'status': 'bad', 'error': "missing post data"}, status=400)

        try:
            data = self.get_data(request_body=request.body)
        except Exception:
            logger.error(f"impossible to decode body")
            return JsonResponse({'status': 'bad', 'error': "impossible to decode body"}, status=400)

        if data is None:
            logger.error(f"None Data")
            return JsonResponse({'status': 'bad', 'error': "None Data"}, status=400)

        try:
            task = self.import_task(data["meta_info"])
        except Exception:
            logger.error(f"Not founded task {data['meta_info']}")
            return JsonResponse({'status': 'bad', 'error': "Not founded task"}, status=400)

        if task is None:
            logger.error(f"Not founded task is none {data['meta_info']}")
            return JsonResponse({'status': 'bad', 'error': "Not founded task"}, status=400)

        try:
            logger.error(f"Execution of {task} with args={data['args']} kwargs={data['kwargs']}")
            result = task.sync_run(*data["args"], **data["kwargs"])
        except Exception:
            logger.error(f"Error on task execution {task} args={data['args']} kwargs={data['kwargs']}")
            return JsonResponse({'status': 'bad', 'error': "Error on task execution"}, status=400)

        return JsonResponse({'status': 'ok', 'result': result}, status=200)


# @csrf_exempt
#     @require_POST
#     def pushed_tasks(request):
#         if not request.body:
#             return HttpResponse("missing post data", status=400)

#         data = None

#         try:
#             json_body = json.loads(request.body)
#             logger.info(f"json_body ==> {json_body}")
#             logger.info(f'encoding ==> {json_body["message"]["attributes"]["encoding"]}')
#             logger.info(f'data ==> {json_body["message"]["data"]}')
#             # decoded_data = str(
#             #     object=json_body["message"]["data"],
#             #     encoding=json_body["message"]["attributes"]["encoding"],
#             #     errors="strict",
#             # )
#             decoded_data_base4 = base64.b64decode(json_body["message"]["data"])
#             logger.info(f"data decoded base64 ==> {decoded_data_base4}")
#             decoded_data = decoded_data_base4.decode(
#                 encoding=json_body["message"]["attributes"]["encoding"],
#                 errors="strict",
#             )
#             # logger.info(
#             logger.info(f"data decoded ==> {decoded_data}")
#             data = json.loads(decoded_data)
#             logger.info(f"data ==> {data}")

#         except Exception as exc:
#             logger.error(exc)
#             return HttpResponse("impossible to decode body", status=400)

#         if data is None:
#             return HttpResponse("None Data", status=400)

#         task = None
#         try:
#             # equiv. of your `import matplotlib.text as text`
#             task_module = importlib.import_module(data["meta_info"]["module_name"])
#             task = getattr(task_module, data["meta_info"]["func_name"])
#         except Exception as exc:
#             logging.error(exc)

#         task.sync_run(*data["args"], **data["kwargs"])

#         return HttpResponse("asd", status=200)



# from decorators import debug, do_twice
import json
import time
import functools
import logging
import inspect
import datetime
# import os
# import google

# from google.oauth2 import service_account
# from google.cloud import pubsub_v1

from django.conf import settings

from django_i3tasks.queue_manager.google_pubsub import PubSubSystemUtils, get_default_queue_setting


logger = logging.getLogger(__name__)

REGITERED_TASKS = dict()


class PubSubTaskUtils:
    def __init__(
        self,
        system_utils,  # 'utf-32',
        encoding="utf-8",  # 'utf-32',
    ):
        self.system_utils = system_utils
        self.encoding = encoding

    def serialize(self, args, kwargs, meta_info):
        serialized_data = json.dumps(
            {
                "args": args,
                "kwargs": kwargs,
                "meta_info": meta_info,
            }
        )
        binary_serialized_data = serialized_data.encode(
            encoding=self.encoding, errors="strict"
        )
        return binary_serialized_data

    def enqueue(self, serialized_data):
        pub_client = self.system_utils.get_publisher_client()
        topic_name = self.system_utils.get_topic_name()

        future = pub_client.publish(
            topic=topic_name,
            data=serialized_data,
            encoding=self.encoding,
            # pub_time=datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
            pub_time=datetime.datetime.utcnow().isoformat(),
            # **attrs: Union[bytes, str]
        )
        future.result()


class TaskDecorator:
    def __init__(
        self,
        func,
        bind=False,
        project_id=settings.PUBSUB_CONFIG.get("PROJECT_ID", None),
        topic_name=get_default_queue_setting("queue_name", "default"),
        subscription_name=get_default_queue_setting("subscription_name", "default"),
        encoding="utf-8",  # 'utf-32',
    ):
        functools.update_wrapper(self, func)
        self._func = func

        self.pubsub_system_utils = PubSubSystemUtils(
            project_id=project_id,
            topic_name=topic_name,
            subscription_name=subscription_name,
            encoding=encoding,
        )
        self.pubsub_task_utils = PubSubTaskUtils(
            system_utils=self.pubsub_system_utils, encoding=encoding
        )

        self.bind = bind
        # https://docs.python.org/3/library/codecs.html#standard-encodings

        self.module_name = inspect.getmodule(self._func).__name__

        self.func_name = self._func.__name__

        self.get_meta_info()

    def __call__(self, *args, **kwargs):
        return self.sync_run(*args, **kwargs)
        # # self.num_calls += 1
        # # logger.info(f"Call {self.num_calls} of {self._func.__name__!r}")
        # start_time = time.perf_counter()    # 1
        # # logger.info("Something is happening before the function is called.")
        # task_result = self._func(*args, **kwargs)
        # # return func(*args, **kwargs)
        # # logger.info("Something is happening after the function is called.")
        # end_time = time.perf_counter()      # 2
        # run_time = end_time - start_time    # 3
        # logger.info(f"Finished {self._func.__name__!r} in {run_time:.4f} secs")
        # return task_result

    def get_meta_info(self):
        self.meta_info = {
            "bind": self.bind,
            "encoding": self.pubsub_system_utils.encoding,
            "module_name": self.module_name,
            "func_name": self.func_name,
        }
        return self.meta_info

    def serialize(self, *args, **kwargs):
        return self.pubsub_task_utils.serialize(
            args=args, kwargs=kwargs, meta_info=self.get_meta_info()
        )

    def enqueue(self, *args, **kwargs):
        serialized_data = self.serialize(*args, **kwargs)
        self.pubsub_task_utils.enqueue(serialized_data)
        # pub_client = self.get_publisher_client()
        # topic_name = self.get_topic_name()
        # binary_serialized_data = self.serialize(*args, **kwargs)
        # future = pub_client.publish(
        #     topic=topic_name,
        #     data=binary_serialized_data,
        #     encoding=self.encoding,
        #     # pub_time=datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
        #     pub_time=datetime.datetime.utcnow().isoformat(),
        #     # **attrs: Union[bytes, str]
        # )
        # future.result()

    def async_run(self, *args, **kwargs):
        # binary_serialized_data = self.serialize(*args, **kwargs)
        self.enqueue(*args, **kwargs)

        # self._run(*args, **kwargs)

    def sync_run(self, *args, **kwargs):
        self.serialize(*args, **kwargs)
        return self._run(*args, **kwargs)

    def _run(self, *args, **kwargs):
        # self.num_calls += 1
        # logger.info(f"Call {self.num_calls} of {self._func.__name__!r}")
        start_time = time.perf_counter()  # 1
        # logger.info("Something is happening before the function is called.")
        task_result = None
        if self.bind:
            task_result = self._func(task_metadata=self, *args, **kwargs)
        else:
            task_result = self._func(*args, **kwargs)
        # return func(*args, **kwargs)
        # logger.info("Something is happening after the function is called.")
        end_time = time.perf_counter()  # 2
        run_time = end_time - start_time  # 3

        mex = f"Finished {self._func.__name__!r} in {run_time:.4f} secs"

        logger.info(mex)
        return task_result



# setup.py
from setuptools import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='django_i3tasks',
    version='0.0.7',
    description='Django app for manage async tasks by http requests',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sajlx/django-i3tasks',
    author='Ivan Bettarini',
    author_email='ivan.bettarini@gmail.com',
    license='GNU General Public License v3.0',
    packages=['django_i3tasks'],
    zip_safe=False,
    install_requires=[
        'django',
        'djangorestframework',
        'requests',
        'croniter>=2.0.1'
        # 'celery',
        # 'redis',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Framework :: Django'
    ]
)


# import media_kit.urls as media_kit_urls
# from action.urls import urlpatterns as action_url_patterns

from django.contrib import admin
# from django.urls import path, re_path
# from django.conf.urls import include

# from rest_framework import routers

# from api import urls as api_urls

# from . import views
# from core import views as core_views


admin.autodiscover()

# router = routers.DefaultRouter()
# router.register(r'user', views.UserViewSet, 'user viewset')

urlpatterns = []
# urlpatterns = [
#     path(r"", core_views.homeview, name="homeview"),
#     path("robots.txt", core_views.robots_txt),
#     path(r"_probe", core_views.probeview, name="probe"),
#     path(r"healtz", core_views.probeview, name="probe"),
#     path(r"healtz_db", core_views.health_check_db, name="probe"),
#     path(r"_status_probe", core_views.probeview, name="status probe"),
#     path(r'admin/', admin.site.urls),
#     path(r'admin/', include('loginas.urls')),
#     re_path(r'user-avatar/(?P<pk>\d+)/', views.UserAvatarView.as_view(), name='user avatar'),
#     path(r'user-choice/', core_views.UserChoiceView.as_view(), name='user choice'),
#     path(r'media-kit/', include(media_kit_urls)),
#     path(r'action/', include(action_url_patterns)),
#     re_path(r'task_status/(?P<task_id>.+)/', core_views.TaskStatusView.as_view(), name='task status'),
#     path(r'', include(router.urls)),
# ] + \
#     api_urls.urlpatterns
